{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Edg where {

import List;
import Nat;
import HOL;
import Substitution;
import Util;
import Term;
import Fun;

data Hterm a b = HFun a [(Hterm a b)] | HHole deriving (Read, Show);

mergeHTerms ::
  forall a b. (Eq a) => [(Hterm a b)] -> [(Hterm a b)] -> Maybe [(Hterm a b)];
mergeHTerms (HFun f fs : hsl) (HFun g gs : hsr) =
  (if eqop f g
    then (case mergeHTerms fs gs of {
           Nothing -> Nothing;
           Just mhs ->
             (case mergeHTerms hsl hsr of {
               Nothing -> Nothing;
               Just res -> Just (HFun f mhs : res);
             });
         })
    else Nothing);
mergeHTerms (HFun f hs : hsl) (HHole : hsr) =
  (case mergeHTerms hsl hsr of {
    Nothing -> Nothing;
    Just res -> Just (HFun f hs : res);
  });
mergeHTerms (HHole : hsl) (h : hsr) =
  (case mergeHTerms hsl hsr of {
    Nothing -> Nothing;
    Just res -> Just (h : res);
  });
mergeHTerms (uw : ux) [] = Nothing;
mergeHTerms [] (uu : uv) = Nothing;
mergeHTerms [] [] = Just [];

merge :: forall a b. (Eq a) => Hterm a b -> Hterm a b -> Maybe (Hterm a b);
merge hl hr =
  (case mergeHTerms [hl] [hr] of {
    Nothing -> Nothing;
    Just res -> Just (nth res Zero_nat);
  });

mergeX ::
  forall a b.
    (Eq a,
      Eq b) => a -> Hterm b a -> [(Hterm b a, a)] -> Maybe [(Hterm b a, a)];
mergeX x h ((ht, y) : ps) =
  (if eqop x y
    then (case merge h ht of {
           Nothing -> Nothing;
           Just m -> mergeX x m ps;
         })
    else (case mergeX x h ps of {
           Nothing -> Nothing;
           Just pairs -> Just ((ht, y) : pairs);
         }));
mergeX x h [] = Just [];

mergeAll :: forall a b. (Eq a, Eq b) => [(Hterm a b, b)] -> Bool;
mergeAll ((h, x) : ps) =
  (case mergeX x h ps of {
    Nothing -> False;
    Just a -> mergeAll a;
  });
mergeAll [] = True;

matchList ::
  forall a b. (Eq a) => [(Hterm a b, Term a b)] -> Maybe [(Hterm a b, b)];
matchList ((HFun f hs, Var x) : ps) =
  (case matchList ps of {
    Nothing -> Nothing;
    Just pairs -> Just ((HFun f hs, x) : pairs);
  });
matchList ((HFun f hs, Fun g ss) : ps) =
  (if eqop f g
    then (case decomp hs ss of {
           Nothing -> Nothing;
           Just pairs -> matchList (append pairs ps);
         })
    else Nothing);
matchList ((HHole, uu) : ps) = matchList ps;
matchList [] = Just [];

matchI :: forall a b. (Eq a, Eq b) => Hterm a b -> Term a b -> Bool;
matchI h t =
  (case matchList [(h, t)] of {
    Nothing -> False;
    Just a -> mergeAll a;
  });

capRM2 ::
  forall a b.
    (Eq a, Eq b) => (a -> [(Term a b, Term a b)]) -> Term a b -> Hterm a b;
capRM2 rm (Fun f ts) =
  let {
    h = HFun f (mapa (capRM2 rm) ts);
  } in (if testPos (\ r -> matchI h (fst r)) (rm f) then HHole else h);
capRM2 uu (Var uv) = HHole;

computeRuleMap ::
  forall a b.
    (Eq a) => [(Term a b, Term a b)] -> Maybe (a -> [(Term a b, Term a b)]);
computeRuleMap ((Var uu, uv) : rules) = Nothing;
computeRuleMap ((Fun f ts, r) : rules) =
  (case computeRuleMap rules of {
    Nothing -> Nothing;
    Just rm -> Just (fun_upd rm f ((Fun f ts, r) : rm f));
  });
computeRuleMap [] = Just (\ f -> []);

capRM ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Hterm a b;
capRM r =
  (case computeRuleMap r of {
    Nothing -> (\ t -> HHole);
    Just a -> capRM2 a;
  });

capRRM ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Hterm a b;
capRRM r = capRM (mapa (\ ra -> (snd ra, fst ra)) r);

checkEdgRM ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 ((Term a b, Term a b), (Term a b, Term a b)) -> Bool;
checkEdgRM r ((s, t), (u, v)) = matchI (capRM r t) u && matchI (capRRM r u) t;

}
